@@ -1182,6 +1182,7 @@ static URI_INLINE UriBool URI_FUNC(OnExitOwnPortUserInfo)(URI_TYPE(ParserState)
  * [ownPortUserInfo]-><_>[ownUserInfo]
  * [ownPortUserInfo]-><~>[ownUserInfo]
  * [ownPortUserInfo]-><->[ownUserInfo]
+ * [ownPortUserInfo]-><@>[ownHost]
  * [ownPortUserInfo]-><NULL>
  */
 static const URI_CHAR * URI_FUNC(ParseOwnPortUserInfo)(URI_TYPE(ParserState) * state, const URI_CHAR * first, const URI_CHAR * afterLast) {
@@ -1206,6 +1207,13 @@ static const URI_CHAR * URI_FUNC(ParseOwnPortUserInfo)(URI_TYPE(ParserState) * s
 	case URI_SET_DIGIT:
 		return URI_FUNC(ParseOwnPortUserInfo)(state, first + 1, afterLast);
 
+	case _UT('@'):
+		state->uri->hostText.afterLast = NULL; /* Not a host, reset */
+		state->uri->portText.first = NULL; /* Not a port, reset */
+		state->uri->userInfo.afterLast = first; /* USERINFO END */
+		state->uri->hostText.first = first + 1; /* HOST BEGIN */
+		return URI_FUNC(ParseOwnHost)(state, first + 1, afterLast);
+
 	default:
 		if (!URI_FUNC(OnExitOwnPortUserInfo)(state, first)) {
 			URI_FUNC(StopMalloc)(state);
@@ -1300,14 +1308,7 @@ static URI_INLINE const URI_CHAR * URI_FUNC(ParsePartHelperTwo)(URI_TYPE(ParserS
 			}
 			afterPathAbsEmpty = URI_FUNC(ParsePathAbsEmpty)(state, afterAuthority, afterLast);
 
-			/* Fix path if only one empty segment */
-			if ((state->uri->pathHead != NULL)
-					&& (state->uri->pathHead->next == NULL)
-					&& (state->uri->pathHead->text.first == state->uri->pathHead->text.afterLast)) {
-				free(state->uri->pathHead);
-				state->uri->pathHead = NULL;
-				state->uri->pathTail = NULL;
-			}
+			URI_FUNC(FixEmptyTrailSegment)(state->uri);
 
 			return afterPathAbsEmpty;
 		}
