@@ -50,12 +50,14 @@ class UriSuite : public Suite {
 
 public:
 	UriSuite() {
+		TEST_ADD(UriSuite::testDistinction)
 		TEST_ADD(UriSuite::testIpFour)
 		TEST_ADD(UriSuite::testIpSixPass)
 		TEST_ADD(UriSuite::testIpSixFail)
 		TEST_ADD(UriSuite::testUri)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort1)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort2)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort22_Bug1948038)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort3)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort4)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort5)
@@ -73,6 +75,7 @@ public:
 		TEST_ADD(UriSuite::testTrailingSlash)
 		TEST_ADD(UriSuite::testAddBase)
 		TEST_ADD(UriSuite::testToString)
+		TEST_ADD(UriSuite::testToString_Bug1950126)
 		TEST_ADD(UriSuite::testToStringCharsRequired)
 		TEST_ADD(UriSuite::testToStringCharsRequired)
 		TEST_ADD(UriSuite::testNormalizeSyntaxMaskRequired)
@@ -82,9 +85,80 @@ public:
 		TEST_ADD(UriSuite::testFilenameUriConversion)
 		TEST_ADD(UriSuite::testCrash_FreeUriMembers_Bug20080116)
 		TEST_ADD(UriSuite::testCrash_MakeOwner_Bug20080207)
+		TEST_ADD(UriSuite::testQueryList)
+		TEST_ADD(UriSuite::testQueryListPair)
 	}
 
 private:
+	bool testDistinctionHelper(const char * uriText, bool expectedHostSet,
+			bool expectedAbsPath, bool expectedEmptyTailSegment) {
+		UriParserStateA state;
+		UriUriA uri;
+		state.uri = &uri;
+
+		int res = uriParseUriA(&state, uriText);
+		if (res != URI_SUCCESS) {
+			uriFreeUriMembersA(&uri);
+			return false;
+		}
+
+		if (expectedHostSet != (uri.hostText.first != NULL)) {
+			uriFreeUriMembersA(&uri);
+			return false;
+		}
+
+		if (expectedAbsPath != (uri.absolutePath == URI_TRUE)) {
+			uriFreeUriMembersA(&uri);
+			return false;
+		}
+
+		if (expectedEmptyTailSegment != ((uri.pathTail != NULL)
+				&& (uri.pathTail->text.first == uri.pathTail->text.afterLast))) {
+			uriFreeUriMembersA(&uri);
+			return false;
+		}
+
+		uriFreeUriMembersA(&uri);
+		return true;
+	}
+
+	void testDistinction() {
+		/*
+============================================================================
+Rule                                | Example | hostSet | absPath | emptySeg
+------------------------------------|---------|---------|---------|---------
+1) URI = scheme ":" hier-part ...   |         |         |         |
+   1) "//" authority path-abempty   | "s://"  | true    |   false |   false
+                                    | "s:///" | true    |   false | true
+   2) path-absolute                 | "s:/"   |   false | true    |   false
+   3) path-rootless                 | "s:a"   |   false |   false |   false
+                                    | "s:a/"  |   false |   false | true
+   4) path-empty                    | "s:"    |   false |   false |   false
+------------------------------------|---------|---------|---------|---------
+2) relative-ref = relative-part ... |         |         |         |
+   1) "//" authority path-abempty   | "//"    | true    |   false |   false
+                                    | "///"   | true    |   false | true
+   2) path-absolute                 | "/"     |   false | true    |   false
+   3) path-noscheme                 | "a"     |   false |   false |   false
+                                    | "a/"    |   false |   false | true
+   4) path-empty                    | ""      |   false |   false |   false
+============================================================================
+		*/
+		TEST_ASSERT(testDistinctionHelper("s://", true, false, false));
+		TEST_ASSERT(testDistinctionHelper("s:///", true, false, true));
+		TEST_ASSERT(testDistinctionHelper("s:/", false, true, false));
+		TEST_ASSERT(testDistinctionHelper("s:a", false, false, false));
+		TEST_ASSERT(testDistinctionHelper("s:a/", false, false, true));
+		TEST_ASSERT(testDistinctionHelper("s:", false, false, false));
+
+		TEST_ASSERT(testDistinctionHelper("//", true, false, false));
+		TEST_ASSERT(testDistinctionHelper("///", true, false, true));
+		TEST_ASSERT(testDistinctionHelper("/", false, true, false));
+		TEST_ASSERT(testDistinctionHelper("a", false, false, false));
+		TEST_ASSERT(testDistinctionHelper("a/", false, false, true));
+		TEST_ASSERT(testDistinctionHelper("", false, false, false));
+	}
+
 	void testIpFour() {
 		URI_TEST_IP_FOUR_FAIL("01.0.0.0");
 		URI_TEST_IP_FOUR_FAIL("001.0.0.0");
@@ -376,6 +450,32 @@ private:
 		uriFreeUriMembersA(&uriA);
 	}
 
+	void testUriUserInfoHostPort22_Bug1948038() {
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		
+		res = uriParseUriA(&stateA, "http://user:21@host/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "user:21", 7 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 7);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+
+		res = uriParseUriA(&stateA, "http://user:1234@192.168.0.1:1234/foo.com");
+		TEST_ASSERT(URI_SUCCESS == res);
+
+		res = uriParseUriA(&stateA, "http://moo:21@moo:21@moo/");
+		TEST_ASSERT(URI_ERROR_SYNTAX == res);
+
+		res = uriParseUriA(&stateA, "http://moo:21@moo:21@moo:21/");
+		TEST_ASSERT(URI_ERROR_SYNTAX == res);
+	}
+
 	void testUriUserInfoHostPort3() {
 		// User info without ":", no port
 		UriParserStateA stateA;
@@ -874,10 +974,11 @@ private:
 		// Port
 		TEST_ASSERT(testToStringHelper(L"http://example.com:123/"));
 		// Path
+		TEST_ASSERT(testToStringHelper(L"http://example.com"));
 		TEST_ASSERT(testToStringHelper(L"http://example.com/"));
 		TEST_ASSERT(testToStringHelper(L"http://example.com/abc/"));
 		TEST_ASSERT(testToStringHelper(L"http://example.com/abc/def"));
-		TEST_ASSERT(testToStringHelper(L"http://example.com/abc/def"));
+		TEST_ASSERT(testToStringHelper(L"http://example.com/abc/def/"));
 		TEST_ASSERT(testToStringHelper(L"http://example.com//"));
 		TEST_ASSERT(testToStringHelper(L"http://example.com/./.."));
 		// Query
@@ -919,6 +1020,24 @@ private:
 		TEST_ASSERT(testToStringHelper(L"f:/.//g"));
 	}
 
+	void testToString_Bug1950126() {
+		UriParserStateW state;
+		UriUriW uriOne;
+		UriUriW uriTwo;
+		const wchar_t * const uriOneString = L"http://e.com/";
+		const wchar_t * const uriTwoString = L"http://e.com";
+		state.uri = &uriOne;
+		TEST_ASSERT(URI_SUCCESS == uriParseUriW(&state, uriOneString));
+		state.uri = &uriTwo;
+		TEST_ASSERT(URI_SUCCESS == uriParseUriW(&state, uriTwoString));
+		TEST_ASSERT(URI_FALSE == uriEqualsUriW(&uriOne, &uriTwo));
+		uriFreeUriMembersW(&uriOne);
+		uriFreeUriMembersW(&uriTwo);
+
+		TEST_ASSERT(testToStringHelper(uriOneString));
+		TEST_ASSERT(testToStringHelper(uriTwoString));
+	}
+
 	bool testToStringCharsRequiredHelper(wchar_t * text) {
 		// Parse
 		UriParserStateW state;
@@ -1247,6 +1366,87 @@ private:
 		TEST_ASSERT(true);
 	}
 
+	void testQueryListHelper(const wchar_t * input, int expectedItemCount) {
+		int res;
+		
+		UriBool spacePlusConversion = URI_TRUE;
+		UriBool normalizeBreaks = URI_FALSE;
+		UriBreakConversion breakConversion = URI_BR_DONT_TOUCH;
+
+		int itemCount;
+		UriQueryListW * queryList;
+		res = uriDissectQueryMallocExW(&queryList, &itemCount,
+				input, input + wcslen(input), spacePlusConversion, breakConversion);
+		TEST_ASSERT(res == URI_SUCCESS);
+		TEST_ASSERT(itemCount == expectedItemCount);
+		TEST_ASSERT((queryList == NULL) == (expectedItemCount == 0));
+
+		if (expectedItemCount != 0) {
+			// First 
+			int charsRequired;
+			res = uriComposeQueryCharsRequiredExW(queryList, &charsRequired, spacePlusConversion, 
+					normalizeBreaks);
+			TEST_ASSERT(res == URI_SUCCESS);
+			TEST_ASSERT(charsRequired >= (int)wcslen(input));
+
+			wchar_t * recomposed = new wchar_t[charsRequired + 1];
+			int charsWritten;
+			res = uriComposeQueryExW(recomposed, queryList, charsRequired + 1,
+					&charsWritten, spacePlusConversion, normalizeBreaks);
+			TEST_ASSERT(res == URI_SUCCESS);
+			TEST_ASSERT(charsWritten <= charsRequired);
+			TEST_ASSERT(charsWritten = (int)wcslen(input) + 1);
+			TEST_ASSERT(!wcscmp(input, recomposed));
+			delete [] recomposed;
+
+			recomposed = NULL;
+			res = uriComposeQueryMallocW(&recomposed, queryList);
+			TEST_ASSERT(res == URI_SUCCESS);
+			TEST_ASSERT(recomposed != NULL);
+			TEST_ASSERT(charsWritten = (int)wcslen(input) + 1);
+			TEST_ASSERT(!wcscmp(input, recomposed));
+			free(recomposed);
+		}
+
+		uriFreeQueryListW(queryList);
+	}
+
+	void testQueryList() {
+		testQueryListHelper(L"one=ONE&two=TWO", 2);
+		testQueryListHelper(L"one=ONE&two=&three=THREE", 3);
+		testQueryListHelper(L"one=ONE&two&three=THREE", 3);
+		testQueryListHelper(L"one=ONE", 1);
+		testQueryListHelper(L"one", 1);
+		testQueryListHelper(L"", 0);
+	}
+
+	void testQueryListPairHelper(const char * pair, const char * unescapedKey,
+			const char * unescapedValue, const char * fixed = NULL) {
+		int res;
+		UriQueryListA * queryList;
+		int itemCount;
+		
+		res = uriDissectQueryMallocA(&queryList, &itemCount, pair, pair + strlen(pair));
+		TEST_ASSERT(res == URI_SUCCESS);
+		TEST_ASSERT(queryList != NULL);
+		TEST_ASSERT(itemCount = 1);
+		TEST_ASSERT(!strcmp(queryList->key, unescapedKey));
+		TEST_ASSERT(!strcmp(queryList->value, unescapedValue));
+
+		char * recomposed;
+		res = uriComposeQueryMallocA(&recomposed, queryList);
+		TEST_ASSERT(res == URI_SUCCESS);
+		TEST_ASSERT(recomposed != NULL);
+		TEST_ASSERT(!strcmp(recomposed, (fixed != NULL) ? fixed : pair));
+		free(recomposed);
+	}
+
+	void testQueryListPair() {
+		testQueryListPairHelper("one+two+%26+three=%2B", "one two & three", "+");
+		testQueryListPairHelper("one=two=three", "one", "two=three", "one=two%3Dthree");
+		testQueryListPairHelper("one=two=three=four", "one", "two=three=four", "one=two%3Dthree%3Dfour");
+	}
+
 };
 
 
